@@ -428,7 +428,7 @@ int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
     fgets (buf, sizeof (buf), msg->fp);
   }
 
-  h->content->length = ftell (msg->fp) - h->content->offset + ctx->hdrs[msgno]->lines;
+  h->content->length = ftell (msg->fp) - h->content->offset;
 
   /* This needs to be done in case this is a multipart message */
 #ifdef HAVE_PGP
